"""Module dedicated to landmarking time-series meta-features."""
import typing as t
import warnings

import numpy as np
import sklearn.model_selection
import sklearn.gaussian_process
import sklearn.base
import statsmodels.tsa.arima.model
import statsmodels.tsa.holtwinters
import statsmodels.tools.sm_exceptions
import statsmodels.base.model

import pymfe._utils as _utils
import pymfe._period as _period
import pymfe._models as _models
import pymfe._embed as _embed

try:
    import pymfe.general as general

except ImportError:
    pass

try:
    import pymfe.autocorr as autocorr

except ImportError:
    pass


class MFETSLandmarking:
    """Extract time-series meta-features from Landmarking group."""

    @classmethod
    def _standard_pipeline_sklearn(
        cls,
        y: np.ndarray,
        model: t.Union[_models.BaseModel, sklearn.base.BaseEstimator],
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        X: t.Optional[np.ndarray] = None,
        args_fit: t.Optional[t.Dict[str, t.Any]] = None,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        scale_range: t.Optional[t.Tuple[float, float]] = (0.0, 1.0),
    ) -> np.ndarray:
        """Fit a model using a canonical pipeline with models from sklearn.

        In this pipeline, each instance of the time-series `y[i]` will be
        automatically associated with a value `X[i]` linearly spaced in the
        [0, 1] range if no `X` is given.

        Let `min` and `max` be the values in the first and second element of
        ``scale_range`` argument, respectively.  Therefore, `X[i]` is given
        by:
        $$
            X[i] = min + (max - min) * i / (len(y) - 1)`
        $$

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at
        each iteration of the validation, a distinct single fold is used as
        the test set, and all other previous folds are used as the train set.

        Parameters
        ----------
        y : :obj:`np.ndarray`
            One-dimensional time-series values.

        model : :obj:`_models.BaseModel` or :obj:`sklearn.base.BaseEstimator`
            A sklearn model, or a custom model from the `_models.py` module.
            Must have `.fit(X, y)` and `.predict(X)` methods implemented.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        X : :obj:`np.ndarray`, optional
            Time-stamps for each time-series observation.

        args_fit : dict, optional
            kwargs for fittig the model (`.fit(X, y, **args_fit)`).

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``y``) are used.

        scale_range : tuple of float, optional (default=(0.0, 1.0))
            Range of ``X`` and ``y`` after normalization. If None, then use
            the original scale of ``y``, and `X = [0, 1, 2, ..., len(y) - 1]`.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.
        """
        if tskf is None:
            tskf = sklearn.model_selection.TimeSeriesSplit(
                n_splits=num_cv_folds
            )

        if args_fit is None:
            args_fit = {}

        # Note: 'X' are the unitless timesteps of the timeseries
        if X is None:
            y = _utils.sample_data(ts=y, lm_sample_frac=lm_sample_frac)
            X_range = scale_range if scale_range is not None else (0, y.size)
            X = np.linspace(*X_range, y.size).reshape(-1, 1)

        else:
            y, X = _utils.sample_data(ts=y, X=X, lm_sample_frac=lm_sample_frac)

        y = y.reshape(-1, 1)
        res = np.zeros(tskf.n_splits, dtype=float)

        if scale_range is not None:
            scaler = sklearn.preprocessing.MinMaxScaler(
                feature_range=scale_range
            )

        for ind_fold, (inds_train, inds_test) in enumerate(tskf.split(X)):
            X_train, X_test = X[inds_train, :], X[inds_test, :]
            y_train, y_test = y[inds_train], y[inds_test]

            if scale_range is not None:
                y_train = scaler.fit_transform(y_train).ravel()
                y_test = scaler.transform(y_test).ravel()

            try:
                model.fit(X_train, y_train, **args_fit)
                y_pred = model.predict(X_test).ravel()
                res[ind_fold] = score(y_pred, y_test)

            except (TypeError, ValueError):
                res[ind_fold] = np.nan

        return res

    @classmethod
    def _standard_pipeline_statsmodels(
        cls,
        ts: np.ndarray,
        model_callable: statsmodels.base.model.Model,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        args_inst: t.Optional[t.Dict[str, t.Any]] = None,
        args_fit: t.Optional[t.Dict[str, t.Any]] = None,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        scale_range: t.Optional[t.Tuple[float, float]] = (0.0, 1.0),
    ) -> np.ndarray:
        """Fit a model using a canonical pipeline with models from statsmodels.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at
        each iteration of the validation, a distinct single fold is used as
        the test set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        model_callable : :obj:`statsmodels.base.model.Model`
            Callable model from statsmodels package.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        args_inst : dict, optional
            Extra arguments for the model instantiation.

        args_fit : dict, optional
            Extra arguments for the fit method.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        scale_range : tuple of float, optional (default=(0.0, 1.0))
            Range of ``ts`` after normalization. If None, then use the original
            ``ts``.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.
        """
        if args_inst is None:
            args_inst = {}

        if args_fit is None:
            args_fit = {}

        ts = _utils.sample_data(ts=ts, lm_sample_frac=lm_sample_frac)

        if tskf is None:
            tskf = sklearn.model_selection.TimeSeriesSplit(
                n_splits=num_cv_folds
            )

        ts = ts.reshape(-1, 1)
        res = np.zeros(tskf.n_splits, dtype=float)

        if scale_range is not None:
            scaler = sklearn.preprocessing.MinMaxScaler(
                feature_range=scale_range
            )

        with warnings.catch_warnings():
            # Note: We are ignoring these warnings because they are related to
            # poor model fits when the time-series does not match the algorithm
            # assumptions. This is expected for a couple of model types, since
            # the same unprocessed time-series is fitted in a large amount of
            # different model types. It is also expected that bad score relates
            # with poor fits, and hence the metafeature will still reflect the
            # true (bad) relationship between the model and the data.
            warnings.filterwarnings(
                "ignore",
                module="statsmodels",
                category=statsmodels.tools.sm_exceptions.ConvergenceWarning,
            )

            warnings.filterwarnings(
                "ignore", module="statsmodels", category=RuntimeWarning
            )

            warnings.filterwarnings(
                "ignore",
                module="statsmodels",
                category=statsmodels.tools.sm_exceptions.HessianInversionWarning,
            )

            for ind_fold, (inds_train, inds_test) in enumerate(tskf.split(ts)):
                ts_train, ts_test = ts[inds_train], ts[inds_test]

                if scale_range is not None:
                    ts_train = scaler.fit_transform(ts_train).ravel()
                    ts_test = scaler.transform(ts_test).ravel()

                try:
                    model = model_callable(ts_train, **args_inst).fit(
                        **args_fit
                    )

                    ts_pred = model.predict(
                        start=ts_train.size,
                        end=ts_train.size + ts_test.size - 1,
                    ).ravel()

                    res[ind_fold] = score(ts_pred, ts_test)

                except (TypeError, ValueError):
                    res[ind_fold] = np.nan

        return res

    @classmethod
    def _model_acf_first_nonpos(
        cls,
        ts: np.ndarray,
        perf_ft_method: t.Callable[..., np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        adjusted: bool = True,
        max_nlags: t.Optional[int] = None,
        **kwargs
    ) -> np.ndarray:
        """First non-positive autocorrelation lag of cross-validation erros.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        perf_ft_method : callable
            Internal cross-validation method.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        kwargs:
            Additional parameters for ``perf_ft_method``.

        Returns
        -------
        :obj:`np.ndarray`
            Lag of the first non-positive in the autocorrelation function for
            each forward chaining cross-validation fold.
        """

        def score(ts_pred: np.ndarray, ts_test: np.ndarray) -> float:
            """Score function: autocorrelation of the errors."""
            return autocorr.MFETSAutocorr.ft_acf_first_nonpos(
                ts=ts_pred - ts_test, adjusted=adjusted, max_nlags=max_nlags
            )

        model_acf_first_nonpos = perf_ft_method(
            ts=ts,
            tskf=tskf,
            score=score,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
            **kwargs
        )

        return model_acf_first_nonpos

    @classmethod
    def ft_model_mean(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the global mean forecasting model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at
        each iteration of the validation, a distinct single fold is used as
        the test set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (0, 0, 0)}
        args_fit = {
            "trend": "c",
            "disp": False,
            "transparams": True,
            "maxiter": 1,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_loc_mean(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        loc_prop: float = 0.25,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the local mean forecasting model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at
        each iteration of the validation, a distinct single fold is used as
        the test set, and all other previous folds are used as the train set.

        The local mean model uses the mean of the most recent observation
        as the forecasting value for all the next observations.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        loc_prop : float, optional (default=0.25)
            Fraction of the most recent observations to be used as the training
            data. Must be in (0, 1] range.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = _models.TSLocalMean(train_prop=loc_prop)

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_loc_median(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        loc_prop: float = 0.25,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the local median forecasting model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at
        each iteration of the validation, a distinct single fold is used as
        the test set, and all other previous folds are used as the train set.

        The local median model uses the median of the most recent observation
        as the forecasting value for all the next observations.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        loc_prop : float, optional (default=0.25)
            Fraction of the most recent observations to be used as the training
            data. Must be in (0, 1] range.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = _models.TSLocalMedian(train_prop=loc_prop)

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_sine(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        opt_initial_guess: bool = True,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        random_state: t.Optional[int] = None,
    ) -> np.ndarray:
        """Cross-validated performance of the sine forecasting model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        The sine model tries to optimize parameters such that a function `f` in
        the form `f(x) = a * sin(w * x + b) + c` describes well the time-series
        values.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        opt_initial_guess : bool, optional (default=True)
            If True, make and optimal initial guess. If False, make a faster
            but naive initial guess.

        loc_prop : float, optional (default=0.25)
            Fraction of the most recent observations to be used as the training
            data. Must be in (0, 1] range.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        random_state : int, optional
            Random seed to perform a random naive initial guess. Used only
            if ``opt_initial_guess`` is False.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = _models.TSSine(
            opt_initial_guess=opt_initial_guess, random_state=random_state
        )

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_exp(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the exponential forecasting model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        The exponential model tries to optimize parameters such that a function
        `f` in the form `f(x) = a * exp(b * x) + c` describes well the
        time-series values.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        loc_prop : float, optional (default=0.25)
            Fraction of the most recent observations to be used as the training
            data. Must be in (0, 1] range.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = _models.TSExp()

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_gaussian(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        random_state: t.Optional[int] = None,
    ) -> np.ndarray:
        """Cross-validated performance of the gaussian process model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        loc_prop : float, optional (default=0.25)
            Fraction of the most recent observations to be used as the training
            data. Must be in (0, 1] range.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        random_state : int, optional
            Random seed, to keep the optimization process deterministic.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = sklearn.gaussian_process.GaussianProcessRegressor(
            copy_X_train=False, random_state=random_state
        )

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_linear(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the linear model in time domain.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.
        """
        model = sklearn.linear_model.LinearRegression()

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_linear_embed(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        emb_dim: t.Optional[int] = None,
        lag: t.Optional[t.Union[str, int]] = None,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        emb_dim_cao_e1: t.Optional[np.ndarray] = None,
        emb_dim_cao_e2: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Fit a linear model on the embedded time-series.

        The appropriate lag is estimated using either the detrended time-series
        automutual information function or the autocorrelation function.

        The appropriate dimension is estimated using the Cao's algorithm.

        The time-series is detrended using the Friedman's Super Smoorther
        algorithm.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        emb_bim : int, optional
            Embedding dimension. If not given,

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        loc_prop : float, optional (default=0.25)
            Fraction of the most recent observations to be used as the training
            data. Must be in (0, 1] range.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        lag : int or str, optional (default = None)
            If scalar, return its own value casted to integer,

            If string, it must be one value in {`ami`, `acf`, `acf-nonsig`},
            which defines the strategy of defining the appropriate lag of
            the embedding.
                1. `ami`: uses the first minimum lag of the automutual
                    information of the time-series.
                2. `acf`: uses the first negative lag of the autocorrelation of
                    the time-series.
                3. `acf-nonsig` (default): uses the first non-significant lag
                    of the time-series autocorrelation function.
                    The non-significant value is defined as the first lag that
                    has the absolute value of is autocorrelation below the
                    critical value defined as 1.96 / sqrt(ts.size).

            If None, the lag will be searched will the 'acf-nonsig' criteria.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, then it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``emb_dim`` is
            None and ``lag`` is any of `acf`, `acf-nonsig` or None.  If this
            argument is not given and the previous condiditon is meet, the
            autocorrelation function will be calculated inside this method up
            to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``emb_dim`` is None and ``lag`` is
            `ami`. If not given and the previous condiditon is meet, the
            automutual information function will be calculated inside this
            method up to ``max_nlags``.

        emb_dim_cao_e1 : :obj:`np.ndarray`, optional
            E1 values from the Cao's method. Used to take advantage of
            precomputations. Used only if ``emb_dim`` is None.

        emb_dim_cao_e2 : :obj:`np.ndarray`, optional
            E2 values from the Cao's method. Used to take advantage of
            precomputations. Used only if ``emb_dim`` is None.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Liangyue Cao, Practical method for determining the minimum
            embedding dimension of a scalar time series, Physica D: Nonlinear
            Phenomena, Volume 110, Issues 1–2, 1997, Pages 43-50,
            ISSN 0167-2789, https://doi.org/10.1016/S0167-2789(97)00118-8.
        .. [2] Friedman, J. H. 1984, A variable span scatterplot smoother
            Laboratory for Computational Statistics, Stanford University
            Technical Report No. 5. Available at:
            https://www.slac.stanford.edu/pubs/slacpubs/3250/slac-pub-3477.pdf
            Accessed on May 12 2020.
        .. [3] Fraser AM, Swinney HL. Independent coordinates for strange
            attractors from mutual information. Phys Rev A Gen Phys.
            1986;33(2):1134‐1140. doi:10.1103/physreva.33.1134
        .. [4] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = sklearn.linear_model.LinearRegression()

        ts_scaled = _utils.standardize_ts(ts=ts, ts_scaled=ts_scaled)

        lag = _embed.embed_lag(
            ts=ts_scaled,
            lag=lag,
            max_nlags=max_nlags,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
        )

        if emb_dim is None:
            emb_dim = general.MFETSGeneral.ft_emb_dim_cao(
                ts=ts,
                lag=lag,
                emb_dim_cao_e1=emb_dim_cao_e1,
                emb_dim_cao_e2=emb_dim_cao_e2,
            )

        X = _embed.embed_ts(ts=ts, dim=emb_dim, lag=lag)

        res = cls._standard_pipeline_sklearn(
            y=X[:, 0],
            X=X[:, 1:],
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_linear_seasonal(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        ts_period: t.Optional[int] = None,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the seasonal linear model.

        This model is evaluated using a linear regression of the dummy
        variables marking the corresponding value from the previous season
        onto the time-series values.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        ts_period : int, optional
            Time-series period. If not given, it will be estimated using
            the minima of the absolute autocorrelation function from lag
            1 up to half the time-series size.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        _ts_period = _period.get_ts_period(ts=ts, ts_period=ts_period)

        if _ts_period <= 1:
            raise ValueError("Time-series is not seasonal (period <= 1).")

        X = np.tile(np.arange(_ts_period), 1 + ts.size // _ts_period)[
            : ts.size, np.newaxis
        ]

        # Note: remove one dummy variable to avoid the 'dummy
        # variable trap'.
        X = sklearn.preprocessing.OneHotEncoder(
            sparse=False, drop="first"
        ).fit_transform(X)

        model = sklearn.linear_model.LinearRegression()

        res = cls._standard_pipeline_sklearn(
            y=ts,
            X=X,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_naive(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the naive model.

        The naive model uses only the last observation of the time-series as
        the forecast value for all future timesteps.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = _models.TSNaive()

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_naive_drift(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the naive model with drift.

        The naive model uses the last observation of the time-series as the
        forecast value for all future timesteps, with a drift added based on
        the timestamp. The drift is the angular coefficient from the line that
        crosses the first and the last observation of the time-series (which
        is equivalent to the mean first-order difference of the time-series).

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model = _models.TSNaiveDrift()

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_naive_seasonal(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        ts_period: t.Optional[int] = None,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of the seasonal linear model.

        This model is evaluated using a linear regression of the dummy
        variables marking the corresponding value from the previous season
        onto the time-series values.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        ts_period : int, optional
            Time-series period. If not given, it will be estimated using
            the minima of the absolute autocorrelation function from lag
            1 up to half the time-series size.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        _ts_period = _period.get_ts_period(ts=ts, ts_period=ts_period)

        if _ts_period <= 1:
            raise ValueError("Time-series is not seasonal (period <= 1).")

        model = _models.TSNaiveSeasonal(ts_period=_ts_period)

        res = cls._standard_pipeline_sklearn(
            y=ts,
            model=model,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_arima_100_c(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        solver: str = "lbfgs",
        maxiter: int = 512,
    ) -> np.ndarray:
        """Cross-validated performance of the ARIMA(1,0,0) with constant model.

        ARIMA(1,0,0) is a first-order regressive model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Statistical forecasting: notes on regression and time series
            analysis, Robert Nau, Fuqua School of Business, Duke University.
            URL: https://people.duke.edu/~rnau/411arim.htm#mixed
            Accessed on 26 May 2020.
        .. [2] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (1, 0, 0)}
        args_fit = {
            "trend": "c",
            "disp": False,
            "transparams": True,
            "maxiter": maxiter,
            "solver": solver,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_arima_010_c(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        solver: str = "lbfgs",
        maxiter: int = 512,
    ) -> np.ndarray:
        """Cross-validated performance of the ARIMA(0,1,0) with constant model.

        ARIMA(0,1,0) is a random walk model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Statistical forecasting: notes on regression and time series
            analysis, Robert Nau, Fuqua School of Business, Duke University.
            URL: https://people.duke.edu/~rnau/411arim.htm#mixed
            Accessed on 26 May 2020.
        .. [2] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (0, 1, 0)}
        args_fit = {
            "disp": False,
            "trend": "c",
            "transparams": True,
            "maxiter": maxiter,
            "solver": solver,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_arima_110_c(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        solver: str = "lbfgs",
        maxiter: int = 512,
    ) -> np.ndarray:
        """Cross-validated performance of the ARIMA(1,1,0) with constant model.

        ARIMA(1,1,0) is a differenced first-order autoregressive model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Statistical forecasting: notes on regression and time series
            analysis, Robert Nau, Fuqua School of Business, Duke University.
            URL: https://people.duke.edu/~rnau/411arim.htm#mixed
            Accessed on 26 May 2020.
        .. [2] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (1, 1, 0)}
        args_fit = {
            "disp": False,
            "trend": "c",
            "transparams": True,
            "maxiter": maxiter,
            "solver": solver,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_arima_011_nc(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        solver: str = "lbfgs",
        maxiter: int = 512,
    ) -> np.ndarray:
        """Cross-validated performance of the ARIMA(0,1,1) (no constant) model.

        ARIMA(0,1,1) without constant is a simple exponential smoothing model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Statistical forecasting: notes on regression and time series
            analysis, Robert Nau, Fuqua School of Business, Duke University.
            URL: https://people.duke.edu/~rnau/411arim.htm#mixed
            Accessed on 26 May 2020.
        .. [2] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (0, 1, 1)}
        args_fit = {
            "disp": False,
            "trend": "nc",
            "transparams": True,
            "maxiter": maxiter,
            "solver": solver,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_arima_011_c(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        solver: str = "lbfgs",
        maxiter: int = 512,
    ) -> np.ndarray:
        """Cross-validated performance of the ARIMA(0,1,1) with constant model.

        ARIMA(0,1,1) with constant is a simple exponential smoothing model with
        growth.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Statistical forecasting: notes on regression and time series
            analysis, Robert Nau, Fuqua School of Business, Duke University.
            URL: https://people.duke.edu/~rnau/411arim.htm#mixed
            Accessed on 26 May 2020.
        .. [2] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (0, 1, 1)}
        args_fit = {
            "disp": False,
            "trend": "c",
            "transparams": True,
            "maxiter": maxiter,
            "solver": solver,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_arima_021_c(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        solver: str = "lbfgs",
        maxiter: int = 512,
    ) -> np.ndarray:
        """Cross-validated performance of the ARIMA(0,2,1) with constant model.

        ARIMA(0,2,1) with constant is a linear exponential smoothing model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Statistical forecasting: notes on regression and time series
            analysis, Robert Nau, Fuqua School of Business, Duke University.
            URL: https://people.duke.edu/~rnau/411arim.htm#mixed
            Accessed on 26 May 2020.
        .. [2] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (0, 2, 1)}
        args_fit = {
            "disp": False,
            "trend": "c",
            "transparams": True,
            "maxiter": maxiter,
            "solver": solver,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_arima_112_nc(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        solver: str = "lbfgs",
        maxiter: int = 512,
    ) -> np.ndarray:
        """Cross-validated performance of the ARIMA(1,1,2) (no constant) model.

        ARIMA(1,1,2) without constant is a damped-trend linear exponential
        smoothing model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Statistical forecasting: notes on regression and time series
            analysis, Robert Nau, Fuqua School of Business, Duke University.
            URL: https://people.duke.edu/~rnau/411arim.htm#mixed
            Accessed on 26 May 2020.
        .. [2] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        model_callable = statsmodels.tsa.arima.model.ARIMA

        args_inst = {"order": (1, 1, 2)}
        args_fit = {
            "disp": False,
            "trend": "nc",
            "transparams": True,
            "maxiter": maxiter,
            "solver": solver,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            args_fit=args_fit,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_ses(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Cross-validated performance of a Single Exponential Smoothing model.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        solver : str, optional (default="lbfgs")
            Solver used to optimize the model.

        maxiter : int, optional (default=512)
            Maximum number of optimization iterations.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        args_inst = {
            "initialization_method": "estimated",
        }

        model_callable = statsmodels.tsa.holtwinters.SimpleExpSmoothing

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            args_inst=args_inst,
            score=score,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_hwes_ada(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        ts_period: t.Optional[int] = None,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Performance of a HW(A_{d},A) model.

        Either components (Trend and Seasonal) are additive, and the model has
        a damping component in the trend component.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        ts_period : int, optional
            Time-series period. If not given, it will be estimated using
            the minima of the absolute autocorrelation function from lag
            1 up to half the time-series size.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        .. [2] Holt, C. E. (1957). Forecasting seasonals and trends by
            exponentially weighted averages (O.N.R. Memorandum No. 52).
            Carnegie Institute of Technology, Pittsburgh USA.
            https://doi.org/10.1016/j.ijforecast.2003.09.015
        .. [3] Winters, P. R. (1960). Forecasting sales by exponentially
            weighted moving averages. Management Science, 6, 324–342.
            https://doi.org/10.1287/mnsc.6.3.324
        """
        _ts_period = _period.get_ts_period(ts=ts, ts_period=ts_period)

        if _ts_period <= 1:
            raise ValueError("Time-series is not seasonal (period <= 1).")

        model_callable = statsmodels.tsa.holtwinters.ExponentialSmoothing

        args_inst = {
            "seasonal_periods": _ts_period,
            "trend": "add",
            "seasonal": "add",
            "damped_trend": True,
            "initialization_method": "estimated",
        }

        args_fit = {
            "use_brute": False,
        }

        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            score=score,
            args_inst=args_inst,
            args_fit=args_fit,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
        )

        return res

    @classmethod
    def ft_model_hwes_adm(
        cls,
        ts: np.ndarray,
        score: t.Callable[[np.ndarray, np.ndarray], np.ndarray],
        ts_period: t.Optional[int] = None,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
    ) -> np.ndarray:
        """Performance of a HW(A_{d},M) model.

        The trend component is additive, and the seasonal component is
        multiplicative. Also, the trend component is damped.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        score : callable
            Score function. Must receive two numeric values as the first two
            arguments, and return a single numeric value.

        ts_period : int, optional
            Time-series period. If not given, it will be estimated using
            the minima of the absolute autocorrelation function from lag
            1 up to half the time-series size.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        Returns
        -------
        :obj:`np.ndarray`
            The model performance for each iteration of the cross-validation.

        References
        ----------
        .. [1] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        .. [2] Holt, C. E. (1957). Forecasting seasonals and trends by
            exponentially weighted averages (O.N.R. Memorandum No. 52).
            Carnegie Institute of Technology, Pittsburgh USA.
            https://doi.org/10.1016/j.ijforecast.2003.09.015
        .. [3] Winters, P. R. (1960). Forecasting sales by exponentially
            weighted moving averages. Management Science, 6, 324–342.
            https://doi.org/10.1287/mnsc.6.3.324
        """
        _ts_period = _period.get_ts_period(ts=ts, ts_period=ts_period)

        if _ts_period <= 1:
            raise ValueError("Time-series is not seasonal (period <= 1).")

        model_callable = statsmodels.tsa.holtwinters.ExponentialSmoothing

        args_inst = {
            "seasonal_periods": _ts_period,
            "trend": "add",
            "seasonal": "mul",
            "damped_trend": True,
            "initialization_method": "estimated",
        }

        args_fit = {
            "use_brute": False,
        }

        # Note: scaling time-series in [1, 2] interval rather than
        # [0, 1] because for Multiplicative Exponential Models,
        # the time-series values must be strictly positive.
        res = cls._standard_pipeline_statsmodels(
            ts=ts,
            model_callable=model_callable,
            score=score,
            args_inst=args_inst,
            args_fit=args_fit,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
            scale_range=(1, 2),
        )

        return res

    @classmethod
    def ft_model_mean_acf_first_nonpos(
        cls,
        ts: np.ndarray,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        adjusted: bool = True,
        max_nlags: t.Optional[int] = None,
    ) -> np.ndarray:
        """First non-positive autocorrelation lags for Mean model errors.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set. Then, the
        autocorrelation function is calculated from the test set errors (i.e.
        true values subtracted by the predicted values).

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        Returns
        -------
        :obj:`np.ndarray`
            First non-positive autocorrelation function lag of the mean model
            errors, for each cross-validation fold.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        acf_first_nonpos_mean = cls._model_acf_first_nonpos(
            ts=ts,
            perf_ft_method=cls.ft_model_mean,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
            adjusted=adjusted,
            max_nlags=max_nlags,
        )

        return acf_first_nonpos_mean

    @classmethod
    def ft_model_linear_acf_first_nonpos(
        cls,
        ts: np.ndarray,
        tskf: t.Optional[sklearn.model_selection.TimeSeriesSplit] = None,
        num_cv_folds: int = 5,
        lm_sample_frac: float = 1.0,
        adjusted: bool = True,
        max_nlags: t.Optional[int] = None,
    ) -> np.ndarray:
        """First non-positive autocorrelation lags for Linear model errors.

        The linear model is in the time domain, i.e., the time-series value
        is regressed onto the timestamps.

        The model score is validated using Forward Chaining, i.e., the full
        time-series is split into ``num_cv_folds`` of equal sizes, and at each
        iteration of the validation, a distinct single fold is used as the test
        set, and all other previous folds are used as the train set. Then, the
        autocorrelation function is calculated from the test set errors (i.e.
        true values subtracted by the predicted values).

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        tskf : :obj:`sklearn.model_selection.TimeSeriesSplit`, optional
            Custom Forward Chaining cross-validatior.

        num_cv_folds : int, optional (default=5)
            Number of test folds. Used only if ``tskf`` is None.

        lm_sample_frac : float, optional (default=1.0)
            Fraction of dataset to use. Default is to use the entire dataset.
            Must be a value in (0, 1] range. Only the most recent time-series
            observations (in the highest indices of ``ts``) are used.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        Returns
        -------
        :obj:`np.ndarray`
            First non-positive autocorrelation function lag of the linear model
            errors, for each cross-validation fold.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        .. [3] Hyndman, R.J., & Athanasopoulos, G. (2018) Forecasting:
            principles and practice, 2nd edition, OTexts: Melbourne,
            Australia. OTexts.com/fpp2. Accessed on 26 May 2020.
        """
        acf_first_nonpos_linear = cls._model_acf_first_nonpos(
            ts=ts,
            perf_ft_method=cls.ft_model_linear,
            tskf=tskf,
            num_cv_folds=num_cv_folds,
            lm_sample_frac=lm_sample_frac,
            adjusted=adjusted,
            max_nlags=max_nlags,
        )

        return acf_first_nonpos_linear



"""Module dedicated to autocorrelation time-series meta-features."""
import typing as t

import statsmodels.tsa.stattools
import numpy as np
import sklearn.gaussian_process

import pymfe._embed as _embed
import pymfe._utils as _utils
import pymfe._detrend as _detrend

try:
    import pymfe.stat_tests as stat_tests

except ImportError:
    pass


class MFETSAutocorr:
    """Extract time-series meta-features from Autocorr group."""

    @classmethod
    def precompute_detrended_acf(
        cls,
        ts: np.ndarray,
        nlags: t.Optional[int] = None,
        adjusted: bool = True,
        **kwargs
    ) -> t.Dict[str, np.ndarray]:
        """Precompute the detrended autocorrelation function.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``detrended_acfs`` (:obj:`np.ndarray`): the autocorrelation
                    function from the detrended time-series.
        """
        precomp_vals = {}

        if "detrended_acfs" not in kwargs:
            precomp_vals["detrended_acfs"] = cls.ft_acf_detrended(
                ts=ts, nlags=nlags, adjusted=adjusted
            )

        return precomp_vals

    @classmethod
    def precompute_gaussian_model(
        cls, ts: np.ndarray, random_state: t.Optional[int] = None, **kwargs
    ) -> t.Dict[str, t.Any]:
        """Precompute a gaussian process model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        random_state : int, optional
            Random seed to optimize the gaussian process model, to keep
            the results reproducible.

        kwargs:
            Additional arguments and previous precomputed items. May
            speed up this precomputation.

        Returns
        -------
        dict
            The following precomputed item is returned:
                * ``gaussian_model`` (:obj:`GaussianProcessRegressor`):
                    Gaussian process fitted model.
                * ``gaussian_resid`` (:obj:`np.ndarray`): Gaussian process
                    model residuals (diference from the original time-series).

            The following item is necessary and, therefore, also precomputed
            if necessary:
                * ``ts_scaled`` (:obj:`np.ndarray`): standardized time-series
                    values (z-score).
        """
        precomp_vals = {}  # type: t.Dict[str, t.Any]

        ts_scaled = kwargs.get("ts_scaled")

        if ts_scaled is None:
            precomp_vals["ts_scaled"] = _utils.standardize_ts(ts=ts)
            ts_scaled = precomp_vals["ts_scaled"]

        if "gaussian_model" not in kwargs:
            gaussian_model = _utils.fit_gaussian_process(
                ts=ts, ts_scaled=ts_scaled, random_state=random_state
            )
            precomp_vals["gaussian_model"] = gaussian_model

        gaussian_model = kwargs.get(
            "gaussian_model", precomp_vals["gaussian_model"]
        )

        if "gaussian_resid" not in kwargs:
            gaussian_resid = _utils.fit_gaussian_process(
                ts=ts,
                ts_scaled=ts_scaled,
                gaussian_model=gaussian_model,
                return_residuals=True,
            )

            precomp_vals["gaussian_resid"] = gaussian_resid

        return precomp_vals

    @classmethod
    def _calc_acf(
        cls,
        ts: np.ndarray,
        nlags: t.Optional[int] = None,
        adjusted: bool = True,
        detrend: bool = True,
        detrended_acfs: t.Optional[np.ndarray] = None,
        ts_detrended: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Precompute the autocorrelation function.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        detrend : bool, optional (default=True)
            If True, detrend the time-series using Friedman's Super Smoother
            before calculating the autocorrelation function, or the user
            given detrended time-series from ``ts_detrended`` argument.

        detrended_acfs : :obj:`np.ndarray`, optional
            This method's return value. Used to take advantage of
            precomputations.

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. Used only if `detrend` is False.

        Returns
        -------
        :obj:`np.ndarray`
            If `detrend` is True, the autocorrelation function up to `nlags`
            lags of the detrended time-series. If `detrend` is False, the
            autocorrelation function up to `nlags` lags of the time-series.
        """
        if detrended_acfs is not None and (
            nlags is None or detrended_acfs.size == nlags
        ):
            return detrended_acfs

        if detrend and ts_detrended is None:
            try:
                ts_detrended = _detrend.decompose(ts=ts, ts_period=0)[2]

            except ValueError:
                pass

        if ts_detrended is None:
            ts_detrended = ts

        if nlags is None:
            nlags = ts.size // 2

        acf = statsmodels.tsa.stattools.acf(
            ts_detrended, nlags=nlags, adjusted=adjusted, fft=True
        )
        return acf[1:]

    @classmethod
    def _calc_pacf(
        cls,
        ts: np.ndarray,
        nlags: t.Optional[int] = None,
        method: str = "ols-adjusted",
        detrend: bool = True,
        ts_detrended: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Precompute the partial autocorrelation function.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the partial autocorrelation function.

        method : str, optional (default="ols-adjusted")
            Method used to estimate the partial autocorrelations. Check the
            `statsmodels.tsa.stattools.pacf` documentation for the complete
            list of the available methods.

        detrend : bool, optional (default=True)
            If True, detrend the time-series using Friedman's Super Smoother
            before calculating the autocorrelation function, or the user
            given detrended time-series from ``ts_detrended`` argument.

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. Used only if `detrend` is False. If not
            given, the time-series is detrended within this method using
            Friedman's Super Smoother.

        Returns
        -------
        :obj:`np.ndarray`
            If `detrend` is True, the partial autocorrelation function up to
            `nlags` lags of the detrended time-series. If `detrend` is False,
            the autocorrelation function up to `nlags` lags of the time-series.
        """
        if nlags is None:
            nlags = 1 + ts.size // 10

        if detrend and ts_detrended is None:
            try:
                ts_detrended = _detrend.decompose(ts=ts, ts_period=0)[2]

            except ValueError:
                pass

        if ts_detrended is None:
            ts_detrended = ts

        pacf = statsmodels.tsa.stattools.pacf(
            ts_detrended, nlags=nlags, method=method
        )
        return pacf[1:]

    @classmethod
    def _first_acf_below_threshold(
        cls,
        ts: np.ndarray,
        threshold: float,
        abs_acf_vals: bool = False,
        max_nlags: t.Optional[int] = None,
        adjusted: bool = True,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """First autocorrelation lag below a given threshold.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        threshold : float
            The threshold to find the first lag below it.

        abs_acf_vals : bool, optional (default=False)
            If True, avaliate the aboslute value of the autocorrelation
            function.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        detrended_acfs : :obj:`np.ndarray`, optional
            This method's return value. Used to take advantage of
            precomputations.

        Returns
        -------
        int or float
            Lag corresponding to the first autocorrelation function the
            given ``threshold``, if any. Return `np.nan` if no such index is
            found.
        """
        detrended_acfs = cls._calc_acf(
            ts=ts,
            nlags=max_nlags,
            adjusted=adjusted,
            detrended_acfs=detrended_acfs,
        )

        if abs_acf_vals:
            # Note: in this case, we are testing if
            # -threshold <= acf <= threshold.
            detrended_acfs = np.abs(detrended_acfs)

        nonpos_acfs = np.flatnonzero(detrended_acfs <= threshold)

        try:
            return nonpos_acfs[0] + 1

        except IndexError:
            return np.nan

    @classmethod
    def ft_acf(
        cls,
        ts: np.ndarray,
        nlags: t.Optional[int] = None,
        adjusted: bool = True,
    ) -> np.ndarray:
        """Autocorrelation function of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        Returns
        -------
        :obj:`np.ndarray`
            The autocorrelation function up to `nlags` lags of the time-series.
        """
        return cls._calc_acf(
            ts=ts, nlags=nlags, adjusted=adjusted, detrend=False
        )

    @classmethod
    def ft_acf_detrended(
        cls,
        ts: np.ndarray,
        nlags: t.Optional[int] = None,
        adjusted: bool = True,
        ts_detrended: t.Optional[np.ndarray] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Autocorrelation function of the detrended time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. If not given, the time-series is detrended
            within this method using Friedman's Super Smoother.

        detrended_acfs : :obj:`np.ndarray`, optional
            This method's return value. Used to take advantage of
            precomputations.

        Returns
        -------
        :obj:`np.ndarray`
            The autocorrelation function up to `nlags` lags of the detrended
            time-series.
        """
        return cls._calc_acf(
            ts=ts,
            nlags=nlags,
            adjusted=adjusted,
            detrend=True,
            detrended_acfs=detrended_acfs,
            ts_detrended=ts_detrended,
        )

    @classmethod
    def ft_acf_diff(
        cls,
        ts: np.ndarray,
        num_diff: int = 1,
        nlags: t.Optional[int] = None,
        detrend: bool = True,
        ts_detrended: t.Optional[np.ndarray] = None,
        adjusted: bool = True,
    ) -> np.ndarray:
        """Autocorrelation function of the differenced time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        num_diff : int, optional (default=1)
            Order of differentiation.

        nlags : int, optional
            Number of lags to calculate the autocorrelation function.

        detrend : bool, optional (default=True)
            If True, detrend the time-series using Friedman's Super Smoother
            before calculating the autocorrelation function, or the user
            given detrended time-series from ``ts_detrended`` argument.

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. If not given and ``detrend`` is True, the
            time-series is detrended within this method using Friedman's Super
            Smoother.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        Returns
        -------
        :obj:`np.ndarray`
            The autocorrelation function up to `nlags` lags of the differenced
            time-series.
        """
        return cls._calc_acf(
            ts=np.diff(ts, n=num_diff),
            detrend=detrend,
            nlags=nlags,
            adjusted=adjusted,
            ts_detrended=ts_detrended,
        )

    @classmethod
    def ft_pacf(
        cls,
        ts: np.ndarray,
        nlags: t.Optional[int] = None,
        method: str = "ols-adjusted",
    ) -> np.ndarray:
        """Partial autocorrelation function of the time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the partial autocorrelation function.

        method : str, optional (default="ols-adjusted")
            Method used to estimate the partial autocorrelations. Check the
            `statsmodels.tsa.stattools.pacf` documentation for the complete
            list of the available methods.

        Returns
        -------
        :obj:`np.ndarray`
            The autocorrelation function up to `nlags` lags of the time-series.
        """
        return cls._calc_pacf(ts=ts, nlags=nlags, method=method, detrend=False)

    @classmethod
    def ft_pacf_detrended(
        cls,
        ts: np.ndarray,
        nlags: t.Optional[int] = None,
        method: str = "ols-adjusted",
        ts_detrended: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Partial autocorrelation function of the detrended time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the partial autocorrelation function.

        method : str, optional (default="ols-adjusted")
            Method used to estimate the partial autocorrelations. Check the
            `statsmodels.tsa.stattools.pacf` documentation for the complete
            list of the available methods.

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. If not given, the time-series is detrended
            within this method using Friedman's Super Smoother.

        Returns
        -------
        :obj:`np.ndarray`
            The partial autocorrelation function up to `nlags` lags of the
            detrended time-series.
        """
        return cls._calc_pacf(
            ts=ts,
            nlags=nlags,
            method=method,
            detrend=True,
            ts_detrended=ts_detrended,
        )

    @classmethod
    def ft_pacf_diff(
        cls,
        ts: np.ndarray,
        num_diff: int = 1,
        nlags: t.Optional[int] = None,
        method: str = "ols-adjusted",
        detrend: bool = True,
        ts_detrended: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Partial autocorrelation function of the differenced time-series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional
            Number of lags to calculate the partial autocorrelation function.

        method : str, optional (default="ols-adjusted")
            Method used to estimate the partial autocorrelations. Check the
            `statsmodels.tsa.stattools.pacf` documentation for the complete
            list of the available methods.

        detrend : bool, optional (default=True)
            If True, detrend the time-series using Friedman's Super Smoother
            before calculating the autocorrelation function, or the user
            given detrended time-series from ``ts_detrended`` argument.

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. Used only if `detrend` is False. If not
            given, the time-series is detrended within this method using
            Friedman's Super Smoother.

        Returns
        -------
        :obj:`np.ndarray`
            If `detrend` is True, the partial autocorrelation function up to
            `nlags` lags of the detrended time-series. If `detrend` is False,
            the autocorrelation function up to `nlags` lags of the time-series.
        """
        return cls._calc_pacf(
            ts=np.diff(ts, n=num_diff),
            nlags=nlags,
            method=method,
            detrend=detrend,
            ts_detrended=ts_detrended,
        )

    @classmethod
    def ft_acf_first_nonsig(
        cls,
        ts: np.ndarray,
        max_nlags: t.Optional[int] = None,
        adjusted: bool = True,
        threshold: t.Optional[t.Union[int, float]] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """First non-significative detrended autocorrelation lag.

        The critical value to determine if a autocorrelation is significative
        is 1.96 / sqrt(len(ts)), but can be changed using the ``threshold``
        parameter.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        threshold : int or float, default
            The critical value to determine if a autocorrelation value is
            significative or not. This means that any autocorrelation with
            absolute value higher than is considered significative. If None,
            then the threshold used will be 1.96 / sqrt(len(ts)).

        ts_detrended : :obj:`np.ndarray`, optional
            Detrended time-series. Used only if `detrend` is False. If not
            given, the time-series is detrended within this method using
            Friedman's Super Smoother.

        Returns
        -------
        int or float
            Lag corresponding to the first autocorrelation with absolute value
            below the given ``threshold``, if any. Return `np.nan` if no such
            index is found.
        """
        if threshold is None:
            threshold = 1.96 / np.sqrt(ts.size)

        res = cls._first_acf_below_threshold(
            ts=ts,
            threshold=threshold,
            abs_acf_vals=True,
            max_nlags=max_nlags,
            adjusted=adjusted,
            detrended_acfs=detrended_acfs,
        )
        return res

    @classmethod
    def ft_acf_first_nonpos(
        cls,
        ts: np.ndarray,
        max_nlags: t.Optional[int] = None,
        adjusted: bool = True,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """First non-positive detrended autocorrelation lag.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        detrended_acfs : :obj:`np.ndarray`, optional
            Detrended time-series autocorrelation function with each index
            corresponding to its lag starting from the lag 1.

        Returns
        -------
        int or float
            Lag corresponding to the first autocorrelation below or equal
            zero, if any. Return `np.nan` if no such index is found.
        """
        res = cls._first_acf_below_threshold(
            ts=ts,
            threshold=0,
            abs_acf_vals=False,
            max_nlags=max_nlags,
            adjusted=adjusted,
            detrended_acfs=detrended_acfs,
        )
        return res

    @classmethod
    def ft_first_acf_locmin(
        cls,
        ts: np.ndarray,
        max_nlags: t.Optional[int] = None,
        adjusted: bool = True,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> t.Union[int, float]:
        """First local minima detrended autocorrelation lag.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        detrended_acfs : :obj:`np.ndarray`, optional
            Detrended time-series autocorrelation function with each index
            corresponding to its lag starting from the lag 1.

        Returns
        -------
        int or float
            Lag corresponding to the first autocorrelation below or equal
            zero, if any. Return `np.nan` if no such index is found.
        """
        detrended_acfs = cls._calc_acf(
            ts=ts,
            nlags=max_nlags,
            adjusted=adjusted,
            detrended_acfs=detrended_acfs,
        )

        acfs_locmin = np.flatnonzero(
            _utils.find_crit_pt(detrended_acfs, type_="min")
        )

        try:
            return acfs_locmin[0] + 1

        except IndexError:
            return np.nan

    @classmethod
    def ft_trev(
        cls,
        ts: np.ndarray,
        lag: t.Optional[t.Union[str, int]] = None,
        only_numerator: bool = False,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
    ) -> float:
        """Normalized nonlinear autocorrelation Trev statistic.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        lag : int or str, optional
            Lag to calculate the statistic. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        only_numerator : bool, optional (default=False)
            If True, return only the numerator from this statistic definition.
            Check `autocorr.MFETSAutocorr.ft_trev` documentation for more
            information.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        Returns
        ------
        float
            Trev statistic.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
        """
        _lag = _embed.embed_lag(
            ts=ts,
            lag=lag,
            max_nlags=max_nlags,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
        )

        diff = ts[_lag:] - ts[:-_lag]

        numen = np.mean(np.power(diff, 3))

        if only_numerator:
            return numen

        denom = np.power(np.mean(np.square(diff)), 1.5)
        trev = numen / denom

        return trev

    @classmethod
    def ft_tc3(
        cls,
        ts: np.ndarray,
        lag: t.Optional[t.Union[str, int]] = None,
        only_numerator: bool = False,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
    ) -> float:
        """Normalized nonlinear autocorrelation Tc3 statistic.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        lag : int or str, optional
            Lag to calculate the statistic. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        only_numerator : bool, optional (default=False)
            If True, return only the numerator from this statistic definition.
            Check `autocorr.MFETSAutocorr.ft_tc3` documentation for more
            information.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        Returns
        ------
        float
            Tc3 statistic.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
        """
        _lag = _embed.embed_lag(
            ts=ts,
            lag=lag,
            max_nlags=max_nlags,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
        )

        ts_shift_1 = ts[: -2 * _lag]
        ts_shift_2 = ts[_lag:-_lag]
        ts_shift_3 = ts[2 * _lag :]

        _aux = ts_shift_1 * ts_shift_2
        numen = np.mean(_aux * ts_shift_3)

        if only_numerator:
            return numen

        denom = np.abs(np.mean(_aux)) ** 1.5

        tc3 = numen / denom

        return tc3

    @classmethod
    def ft_gen_autocorr(
        cls,
        ts: np.ndarray,
        alpha: float = 1,
        beta: float = 1,
        lag: t.Optional[t.Union[str, int]] = None,
        max_nlags: t.Optional[int] = None,
        detrended_acfs: t.Optional[np.ndarray] = None,
        detrended_ami: t.Optional[np.ndarray] = None,
    ) -> float:
        """Generalized autocorrelation of the time-series.

        If alpha = beta, estimates how values of the same order of magnitude
        are related in time. Otherwise, estimates correlations between
        different magnitudes of the time series.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        alpha : float, optional (default=1)
            Non-zero parameter.

        beta : float, optional (default=1)
            Non-zero parameter.

        lag : int or str, optional
            Lag to calculate the statistic. It must be a strictly positive
            value, None or a string in {`acf`, `acf-nonsig`, `ami`}. In the
            last two type of options, the lag is estimated within this method
            using the given strategy method (or, if None, it is used the
            strategy `acf-nonsig` by default) up to ``max_nlags``.
                1. `acf`: the lag corresponds to the first non-positive value
                    in the autocorrelation function.
                2. `acf-nonsig`: lag corresponds to the first non-significant
                    value in the autocorrelation function (absolute value below
                    the critical value of 1.96 / sqrt(ts.size)).
                3. `ami`: lag corresponds to the first local minimum of the
                    time-series automutual information function.

        max_nlags : int, optional
            If ``lag`` is not a numeric value, than it will be estimated using
            either the time-series autocorrelation or mutual information
            function estimated up to this argument value.

        detrended_acfs : :obj:`np.ndarray`, optional
            Array of time-series autocorrelation function (for distinct ordered
            lags) of the detrended time-series. Used only if ``lag`` is any of
            `acf`, `acf-nonsig` or None.  If this argument is not given and the
            previous condiditon is meet, the autocorrelation function will be
            calculated inside this method up to ``max_nlags``.

        detrended_ami : :obj:`np.ndarray`, optional
            Array of time-series automutual information function (for distinct
            ordered lags). Used only if ``lag`` is `ami`. If not given and the
            previous condiditon is meet, the automutual information function
            will be calculated inside this method up to ``max_nlags``.

        Returns
        -------
        float
            Generalized autocorrelation of the time-series.

        References
        ----------
        .. [1] S.M. Duarte Queirós, L.G. Moyano, Yet on statistical properties
            of traded volume: Correlation and mutual information at different
            value magnitudes, Physica A: Statistical Mechanics and its
            Applications, Volume 383, Issue 1, 2007, Pages 10-15, ISSN
            0378-4371, https://doi.org/10.1016/j.physa.2007.04.082.
        .. [2] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [3] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if np.isclose(alpha, 0.0):
            raise ValueError(
                "'alpha' parameter must be nonzero (got {})." "".format(alpha)
            )

        if np.isclose(beta, 0.0):
            raise ValueError(
                "'beta' parameter must be nonzero (got {})." "".format(beta)
            )

        _lag = _embed.embed_lag(
            ts=ts,
            lag=lag,
            max_nlags=max_nlags,
            detrended_acfs=detrended_acfs,
            detrended_ami=detrended_ami,
        )

        ts_abs = np.abs(ts)
        ts_sft_1 = ts_abs[:-_lag]
        ts_sft_2 = ts_abs[_lag:]

        ts_sft_1_a = ts_sft_1 ** alpha
        ts_sft_2_b = ts_sft_2 ** beta

        ts_sft_1_a_mean = np.mean(ts_sft_1_a)
        ts_sft_2_b_mean = np.mean(ts_sft_2_b)

        gen_autocorr = np.mean(
            ts_sft_1_a * ts_sft_2_b
        ) - ts_sft_1_a_mean * ts_sft_2_b_mean / (
            np.sqrt(np.mean(ts_sft_1 ** (2 * alpha)) - ts_sft_1_a_mean ** 2)
            * np.sqrt(np.mean(ts_sft_2 ** (2 * beta)) - ts_sft_2_b_mean ** 2)
        )

        return gen_autocorr

    @classmethod
    def ft_autocorr_crit_pt(
        cls,
        ts: np.ndarray,
        crit_point_type: str = "non-plateau",
        return_lags: bool = True,
        max_nlags: t.Optional[int] = None,
        adjusted: bool = True,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Lags corresponding to minima or maxima of autocorrelation function.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        crit_point_type : str, optional (default="non-plateau")
            Critical point type. Must be a value in {`non-plateau`, `plateau`,
            `min`, `max`, `any`}.

        return_lags : bool, optional (default=True)
            If True, return the lags corresponding to the autocorrelation
            function critical points. If False, return a binary array marking
            with `1` the positions corresponding to the critical points, and
            `0` otherwise.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        detrended_acfs : :obj:`np.ndarray`, optional
            Detrended time-series autocorrelation function with each index
            corresponding to its lag starting from the lag 1.

        Returns
        -------
        :obj:`np.ndarray`
            If `return_lags` is True, return the lags corresponding to the
            autocorrelation function critical points. If `return_lags` is
            False, return a binary array marking with `1` the lag indices
            (starting from lag 1) corresponding to the autocorrelation function
            critical points.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        detrended_acfs = cls._calc_acf(
            ts=ts,
            nlags=max_nlags,
            adjusted=adjusted,
            detrended_acfs=detrended_acfs,
        )

        ac_shape = _utils.find_crit_pt(
            arr=detrended_acfs, type_=crit_point_type
        )

        # Note: in 'hctsa', either the sum or the mean is returned.
        # However, to enable summarization, here we return the whole
        # array.

        if return_lags:
            return np.flatnonzero(ac_shape)

        return ac_shape.astype(int)

    @classmethod
    def ft_gresid_autocorr(
        cls,
        ts: np.ndarray,
        nlags: int = 8,
        adjusted: bool = True,
        random_state: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        gaussian_resid: t.Optional[np.ndarray] = None,
        gaussian_model: t.Optional[
            sklearn.gaussian_process.GaussianProcessRegressor
        ] = None,
    ) -> np.ndarray:
        """Autocorrelation function of the gaussian process model residuals.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional (default=8)
            Number of lags evaluated in the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        random_state : int, optional
            Random seed to optimize the gaussian process model, to keep
            the results reproducible.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations. Used only if ``gaussian_resid`` is None.

        gaussian_resid : :obj:`np.ndarray`, optional
            Residuals of a gaussian process. Used to take advantage of
            precomputations.

        gaussian_model : :obj:`GaussianProcessRegressor`, optional
            A fitted model of a gaussian process. Used to take advantage of
            precomputations. Used only if ``gaussian_resid`` is None.

        Returns
        -------
        :obj:`np.ndarray`
            Autocorrelation function of the gaussian process residuals up
            to ``nlags``.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if gaussian_resid is None:
            gaussian_resid = _utils.fit_gaussian_process(
                ts=ts,
                ts_scaled=ts_scaled,
                random_state=random_state,
                gaussian_model=gaussian_model,
                return_residuals=True,
            )

        gaussian_resid_acf = cls._calc_acf(
            ts=gaussian_resid, nlags=nlags, adjusted=adjusted
        )

        return gaussian_resid_acf

    @classmethod
    def ft_gresid_lbtest(
        cls,
        ts: np.ndarray,
        nlags: int = 8,
        return_pval: bool = True,
        random_state: t.Optional[int] = None,
        ts_scaled: t.Optional[np.ndarray] = None,
        gaussian_resid: t.Optional[np.ndarray] = None,
        gaussian_model: t.Optional[
            sklearn.gaussian_process.GaussianProcessRegressor
        ] = None,
    ) -> np.ndarray:
        """Ljung–Box test in the residuals of a gaussian process model.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        nlags : int, optional (default=8)
            Number of lags evaluated in the Ljung-Box test.

        return_pval : bool, optional (default=True)
            If True, return the p-value of the test instead of the test
            statistic.

        random_state : int, optional
            Random seed to optimize the gaussian process model, to keep
            the results reproducible.

        ts_scaled : :obj:`np.ndarray`, optional
            Standardized time-series values. Used to take advantage of
            precomputations. Used only if ``gaussian_resid`` is None.

        gaussian_resid : :obj:`np.ndarray`, optional
            Residuals of a gaussian process. Used to take advantage of
            precomputations.

        gaussian_model : :obj:`GaussianProcessRegressor`, optional
            A fitted model of a gaussian process. Used to take advantage of
            precomputations. Used only if ``gaussian_resid`` is None.

        Returns
        -------
        :obj:`np.ndarray`
            If `return_pval` is False, Ljung-Box test statistic for each lag
            of the gaussian process residuals.
            If `return_pval` is True, p-value associated with the Ljung-Box
            test statistic for each lag of the gaussian process residuals.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        if gaussian_resid is None:
            gaussian_resid = _utils.fit_gaussian_process(
                ts=ts,
                ts_scaled=ts_scaled,
                random_state=random_state,
                gaussian_model=gaussian_model,
                return_residuals=True,
            )

        gaussian_lb_test = stat_tests.MFETSStatTests.ft_test_lb(
            ts_residuals=gaussian_resid,
            max_nlags=nlags,
            return_pval=return_pval,
        )

        return gaussian_lb_test

    @classmethod
    def ft_autocorr_out_dist(
        cls,
        ts: np.ndarray,
        p: float = 0.8,
        max_nlags: t.Optional[int] = None,
        adjusted: bool = True,
        detrended_acfs: t.Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """Distance between the autocorrelation with and without outliers.

        This method calculates the time-series autocorrelation function
        for all observations, and the aucorrelation function of the
        time-series without a subset of the most extreme values (cut at
        the ``p`` quantile of all absolute values). It is returned the
        absolute difference between these two autocorrelations.

        Parameters
        ----------
        ts : :obj:`np.ndarray`
            One-dimensional time-series values.

        p : float, optional (default=0.8)
            Quantile of cut in the set of the time-series absolute values to
            determine which instances are considered outliers.

        max_nlags : int, optional
            Number of lags to avaluate the autocorrelation function.

        adjusted : bool, optional (default=True)
            If True, the autocorrelation function is corrected for statistical
            bias.

        detrended_acfs : :obj:`np.ndarray`, optional
            Detrended time-series autocorrelation function with each index
            corresponding to its lag starting from the lag 1.

        Returns
        -------
        :obj:`np.ndarray`
            Absolute difference element-wise between each autocorrelation
            with and without outliers.

        References
        ----------
        .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
            for Automated Time-Series Phenotyping Using Massive Feature
            Extraction, Cell Systems 5: 527 (2017).
            DOI: 10.1016/j.cels.2017.10.001
        .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
            time-series analysis: the empirical structure of time series and
            their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
            DOI: 10.1098/rsif.2013.0048
        """
        detrended_acfs = cls._calc_acf(
            ts=ts,
            nlags=max_nlags,
            adjusted=adjusted,
            detrended_acfs=detrended_acfs,
        )

        ts_abs = np.abs(ts)
        ts_inliners = ts[ts_abs <= np.quantile(ts_abs, p)]

        ts_inliners_acfs = cls._calc_acf(
            ts=ts_inliners, nlags=max_nlags, adjusted=adjusted
        )

        dist_acfs = np.abs(
            detrended_acfs[: ts_inliners_acfs.size] - ts_inliners_acfs
        )

        return dist_acfs



"""Utility functions used ubiquitously over this library source code."""
import typing as t
import operator

import sklearn.preprocessing
import sklearn.gaussian_process
import numpy as np
import pandas as pd


def apply_on_tiles(
    ts: np.ndarray,
    num_tiles: int,
    func: t.Callable[[np.ndarray], t.Any],
    *args,
    **kwargs
) -> np.ndarray:
    """Apply a function on time-series tiles (non-overlapping windows)."""
    if num_tiles > 0.5 * ts.size:
        raise ValueError(
            "'num_tiles' ({}) larger than half the "
            "time-series size ({}).".format(num_tiles, 0.5 * ts.size)
        )

    res = np.array(
        [
            func(split, *args, **kwargs)  # type: ignore
            for split in np.array_split(ts, num_tiles)
        ],
        dtype=float,
    )

    return res


def process_window_size(
    ts: np.ndarray, window_size: t.Union[float, int]
) -> int:
    """Standardize rolling window sizes.

    The following steps are made in this method:
        1. Check if window size is stricly positive (otherwise, raise a
            ValueError exception).
        2. If 0 < window_size < 1, transform it to the corresponding integer
            size, based on the ``ts`` length. Otherwise, set window_size to
            min(len(ts), window_size).
        3. Force window_size be an odd value, to keep the reference window
            value at the center of the window.
    """
    if window_size <= 0:
        raise ValueError(
            "'window_size' must be positive (got {})." "".format(window_size)
        )

    if 0 < window_size < 1:
        window_size = max(1, int(np.ceil(window_size * ts.size)))

    else:
        window_size = min(ts.size, window_size)

    if window_size % 2 == 0:
        # Note: forcing 'window_size' be a off number in order to the
        # reference value be exactly on the window center (and avoid
        # possibility of bias towards the larger tail)
        window_size -= 1

    return int(window_size)


def standardize_ts(
    ts: np.ndarray, ts_scaled: t.Optional[np.ndarray] = None
) -> np.ndarray:
    """Standardize (z-score normalization) time-series."""
    if ts_scaled is None:
        if not isinstance(ts, np.ndarray):
            ts = np.asarray(ts, dtype=float)

        if ts.ndim == 1:
            ts = ts.reshape(-1, 1)

        return sklearn.preprocessing.StandardScaler().fit_transform(ts).ravel()

    return ts


def get_rolling_window(
    ts: np.ndarray,
    window_size: t.Union[float, int],
    center: bool = True,
    scale: bool = True,
    ts_scaled: t.Optional[np.ndarray] = None,
    ts_rol_win: t.Optional[pd.core.window.rolling.Rolling] = None,
) -> pd.core.window.rolling.Rolling:
    """Apply a function on time-series rolling (overlapping) windows.

    If ``center`` is True, then each rolling window is centered at the central
    instance rather than the initial instance.

    If ``scaled`` is True, the time-series will be standardized before the
    rolling window is built (or ``ts_scaled`` will be used, if not None).
    """
    window_size = process_window_size(ts=ts, window_size=window_size)

    if ts_rol_win is not None and window_size == ts_rol_win.window:
        return ts_rol_win

    if scale:
        ts = standardize_ts(ts=ts, ts_scaled=ts_scaled)

    return pd.Series(ts).rolling(window_size, center=center)


def sample_data(
    ts: np.ndarray,
    lm_sample_frac: float,
    X: t.Optional[np.ndarray] = None,
) -> t.Union[np.ndarray, t.Tuple[np.ndarray, np.ndarray]]:
    """Select ``lm_sample_frac`` percent of data from ``ts``.

    ``lm_sample_frac`` must be in (0, 1] range.

    The chosen instances are the most recent ones (on the higher indices
    of the ``ts`` array).

    ``X`` is any array of values associated to each ``ts`` observation, and
    will be sampled alongside it (mantaining the index correspondence) if
    given. Tipically, ``X`` is the timestamps of each ``ts`` observation.
    """
    if not 0.0 < lm_sample_frac <= 1.0:
        raise ValueError(
            "'lm_sample_frac' must be in (0, 1] range (got {})."
            "".format(lm_sample_frac)
        )

    threshold = int(np.ceil(ts.size * lm_sample_frac))

    if threshold >= ts.size:
        if X is not None:
            return ts, X

        return ts

    if X is not None:
        return ts[-threshold:], X[-threshold:]

    return ts[-threshold:]


def find_plateau_pt(
    arr: np.ndarray, arr_diff: t.Optional[np.ndarray] = None
) -> np.ndarray:
    """Find plateau points in array.

    ``arr_diff`` is the first-order differenced ``arr``, which can be
    passed to speed up this computation since this value is needed within
    this function.
    """
    if arr_diff is None:
        arr_diff = np.diff(arr)

    arr_diff_2 = np.diff(arr_diff)

    res = np.logical_and(
        np.isclose(arr_diff_2, 0), np.isclose(arr_diff[:-1], 0)
    )

    return np.hstack((False, res, False))


def find_crit_pt(arr: np.ndarray, type_: str) -> np.ndarray:
    """Find critical points on the given values.

    ``type`` must be in {"min", "max", "plateau", "non-plateau", "any"}.
    """
    if arr.size <= 2:
        raise ValueError(
            "Array too small (size {}). Need at least "
            "3 elements.".format(arr.size)
        )

    VALID_TYPES = {"min", "max", "plateau", "non-plateau", "any"}

    if type_ not in VALID_TYPES:
        raise ValueError(
            "'type_' must be in {} (got '{}')." "".format(type_, VALID_TYPES)
        )

    # Note: first discrete derivative
    arr_diff_1 = np.diff(arr)

    if type_ == "plateau":
        return find_plateau_pt(arr, arr_diff_1)

    turning_pt = arr_diff_1[1:] * arr_diff_1[:-1] < 0

    if type_ == "non-plateau":
        return np.hstack((False, turning_pt, False))

    if type_ == "any":
        plat = find_plateau_pt(arr, arr_diff_1)
        turning_pt = np.hstack((False, turning_pt, False))
        res = np.logical_or(turning_pt, plat)
        return res

    # Note: second discrete derivative
    arr_diff_2 = np.diff(arr_diff_1)

    rel = operator.lt if type_ == "max" else operator.gt

    interest_pt = rel(arr_diff_2, 0)
    local_m = np.logical_and(turning_pt, interest_pt)

    return np.hstack((False, local_m, False))


def fit_gaussian_process(
    ts: np.ndarray,
    random_state: t.Optional[int] = None,
    return_residuals: bool = False,
    ts_scaled: t.Optional[np.ndarray] = None,
    gaussian_model: t.Optional[
        sklearn.gaussian_process.GaussianProcessRegressor
    ] = None,
) -> t.Union[np.ndarray, sklearn.gaussian_process.GaussianProcessRegressor]:
    """Fit a Gaussian Process model to the time-series data.

    The fitted model is returned unless ``return_residuals`` is True, which in
    this case the model residuals is returned instead.
    """
    ts_scaled = standardize_ts(ts=ts, ts_scaled=ts_scaled)

    if gaussian_model is None or return_residuals:
        timestamps = np.linspace(0, 1, ts_scaled.size).reshape(-1, 1)

    if gaussian_model is None:
        gaussian_model = sklearn.gaussian_process.GaussianProcessRegressor(
            copy_X_train=False, random_state=random_state
        )

        gaussian_model.fit(X=timestamps, y=ts_scaled)

    if return_residuals:
        return (ts_scaled - gaussian_model.predict(X=timestamps)).ravel()

    return gaussian_model


def calc_ioe_stats(
    ts: np.ndarray,
    funcs: t.Collection[t.Callable[[np.ndarray], float]],
    ts_scaled: t.Optional[np.ndarray] = None,
    step_size: float = 0.05,
    max_it: int = 1024,
    differentiate: bool = False,
) -> np.ndarray:
    """Get statistics using the iterative outlier exclusion strategy.

    In the iterative outlier exclusion, a uniformly spaced set of thresholds
    over the time-series range is build in increasing order. For each threshold
    it is calculated a statistic of the diference of the timestamp values of
    instances larger or equal than the current threshold.

    Parameters
    ----------
    ts : :obj:`np.ndarray`
        One-dimensional time-series values.

    funcs : callable or list of callables
        Callables that extract the statistics from the timestamps. Every
        callable must receive a numeric list of values as the first argument
        and must return a single numeric value.

    ts_scaled : :obj:`np.ndarray`, optional
        Standardized time-series values. Used to take advantage of
        precomputations.

    step_size : float, optional (default=0.05)
        Increase of the outlier threshold in each iteration. Must be a number
        strictly positive.

    max_it : int, optional (default=1024)
        Maximum number of iterations.

    differentiate : bool, optional (default=False)
        If True, differentiate the timestamps before calculating each
        statistic. If False, all statistics will be calculated on the
        raw timestamps.

    Returns
    -------
    :obj:`np.ndarray`
        Array where each row corresponds to a distinct statistisc extracted
        from the timestamps, and each column corresponds to each iteration
        of the iterative outlier inclusive process. If ``funcs`` is a single
        function, then the return value will be flattened to a 1-D array.

    References
    ----------
    .. [1] B.D. Fulcher and N.S. Jones, "hctsa: A Computational Framework
        for Automated Time-Series Phenotyping Using Massive Feature
        Extraction, Cell Systems 5: 527 (2017).
        DOI: 10.1016/j.cels.2017.10.001
    .. [2] B.D. Fulcher, M.A. Little, N.S. Jones, "Highly comparative
        time-series analysis: the empirical structure of time series and
        their methods", J. Roy. Soc. Interface 10(83) 20130048 (2013).
        DOI: 10.1098/rsif.2013.0048
    """
    if step_size <= 0:
        raise ValueError(
            "'step_size' must be positive (got {})." "".format(step_size)
        )
    try:
        if len(funcs) == 0:
            raise ValueError("'funcs' is empty.")

    except TypeError:
        funcs = [funcs]  # type: ignore

    ts_scaled = standardize_ts(ts=ts, ts_scaled=ts_scaled)

    ts_abs = np.abs(ts_scaled)
    min_abs_ts, max_abs_ts = np.quantile(ts_abs, (0, 1))

    step_size *= np.std(ts_abs)
    _max_it = min(max_it, int(np.ceil(max_abs_ts / step_size)))
    ioe_stats = np.zeros((_max_it, len(funcs)))
    threshold = min_abs_ts
    it = 0

    while it < _max_it:
        threshold += step_size
        outlier_tsteps = np.flatnonzero(ts_abs >= threshold)

        if (
            outlier_tsteps.size < 0.02 * ts_scaled.size
            or outlier_tsteps.size <= 1
        ):
            break

        diff_tsteps = np.diff(outlier_tsteps, int(differentiate))

        ioe_stats[it, :] = [func(diff_tsteps) for func in funcs]
        it += 1

    if len(funcs) == 1:
        return ioe_stats.ravel()

    # Note: transposing in order to each statistic be represented by a row,
    # and not a column.
    return ioe_stats.T


def apply_on_samples(
    ts: np.ndarray,
    func: t.Callable[..., float],
    num_samples: int = 128,
    sample_size_frac: float = 0.2,
    random_state: t.Optional[int] = None,
    **kwargs
) -> np.ndarray:
    """Apply ``func`` on time-series random samples.

    The samples are ``num_samples`` consecutive observations from the
    time-series, starting from random positions.

    Parameters
    ----------
    ts : :obj:`np.ndarray`
        One-dimensional time-series values.

    func : callable
        Function to apply on each sample. Must receive an list of numeric
        values as argument and return a single numeric valud.

    num_samples : int, optional (default=128)
        Number of samples. If zero or less, a empty array will be returned.

    sample_size_frac : float, optional (default=0.2)
        Size of each sample as a fraction of the size of the original
        time-series. Must be in (0, 1) range.

    random_state : int, optional
        Random seed to ensure reproducibility.

    kwargs:
        Extra arguments to ``func``.

    Returns
    -------
    :obj:`np.ndarray`
        ``func`` values from each time-series sample.
    """
    if num_samples <= 0:
        return np.empty(0)

    if not 0 < sample_size_frac < 1:
        raise ValueError(
            "'sample_size_frac' must be in (0, 1) "
            "range (got {}).".format(sample_size_frac)
        )

    if random_state is not None:
        np.random.seed(random_state)

    sample_size = int(np.ceil(ts.size * sample_size_frac))
    start_inds = np.random.randint(ts.size - sample_size + 1, size=num_samples)

    res = np.array(
        [
            func(ts[s_ind : s_ind + sample_size], **kwargs)
            for s_ind in start_inds
        ]
    )

    return res


def discretize(
    ts: np.ndarray,
    num_bins: int,
    strategy: str = "equal-width",
    dtype: t.Type = int,
) -> np.ndarray:
    """Discretize a time-series using a histogram.

    Parameters
    ----------
    ts : :obj:`np.ndarray`
        One-dimensional time-series values.

    num_bins : int
        Number of bins in the histogram.

    strategy : {`equal-width`,`equiprobable`}, optional (default="equal-width")
            Strategy used to define the histogram bins. Must be either
            `equal-width` (bins with equal with) or `equiprobable` (bins
            with the same amount of observations within).

    dtype : type, optional (default=int)
        Output type of the discretized time-series.

    Returns
    -------
    :obj:`np.ndarray`
        Discretized time-series with the selected strategy.
    """
    VALID_METHODS = {"equal-width", "equiprobable"}

    if strategy not in VALID_METHODS:
        raise ValueError(
            "'strategy' must be in {} (got {})."
            "".format(VALID_METHODS, strategy)
        )

    if strategy == "equal-width":
        bins = np.histogram(ts, num_bins)[1][:-1]

    elif strategy == "equiprobable":
        bins = np.quantile(ts, np.linspace(0, 1, num_bins + 1)[:-1])

    ts_disc = np.digitize(ts, bins)

    return ts_disc.astype(dtype)
